table.autosummary {
  width: 100%;
}

table.autosummary thead tr th:first-child,
table.autosummary tbody tr td:first-child {
  width: 25%;
  min-width: 25%;
  max-width: 25%;
  text-overflow: ellipsis;
  white-space: nowrap;
  overflow: hidden;
}

table.autosummary .caption-text {
  display: block;
  width: 100%;
  background-color: var(--color-background-border);
}

table.autosummary a.headerlink {
  display: none;
}

.nbinput .prompt,
.nboutput .prompt {
    display: none;
}
